@@ -1,4 +1,5 @@
 # -*- coding: utf-8 -*-
+import time
 
 try:
     import ujson as json
@@ -6,10 +7,10 @@ except ImportError:
     import json
 
 import requests
+from requests.exceptions import HTTPError, ConnectionError, Timeout
 
 try:
     from requests.packages.urllib3 import fields
-
     format_header_param = fields.format_header_param
 except ImportError:
     format_header_param = None
@@ -28,6 +29,10 @@ FILE_URL = None
 CONNECT_TIMEOUT = 3.5
 READ_TIMEOUT = 9999
 
+RETRY_ON_ERROR = False
+RETRY_TIMEOUT = 2
+MAX_RETRIES = 15
+
 CUSTOM_SERIALIZER = None
 
 ENABLE_MIDDLEWARE = False
@@ -47,11 +52,11 @@ def _make_request(token, method_name, method='get', params=None, files=None):
     :param files: Optional files.
     :return: The result parsed to a JSON dictionary.
     """
-    if API_URL is None:
-        request_url = "https://api.telegram.org/bot{0}/{1}".format(token, method_name)
-    else:
+    if API_URL:
         request_url = API_URL.format(token, method_name)
-    
+    else:
+        request_url = "https://api.telegram.org/bot{0}/{1}".format(token, method_name)
+
     logger.debug("Request: method={0} url={1} params={2} files={3}".format(method, request_url, params, files))
     read_timeout = READ_TIMEOUT
     connect_timeout = CONNECT_TIMEOUT
@@ -62,11 +67,52 @@ def _make_request(token, method_name, method='get', params=None, files=None):
             read_timeout = params.pop('timeout') + 10
         if 'connect-timeout' in params:
             connect_timeout = params.pop('connect-timeout') + 10
-    result = _get_req_session().request(
-        method, request_url, params=params, files=files,
-        timeout=(connect_timeout, read_timeout), proxies=proxy)
+        if 'long_polling_timeout' in params:
+            # For getUpdates: the only function with timeout on the BOT API side
+            params['timeout'] = params.pop('long_polling_timeout')
+
+
+    result = None
+    if RETRY_ON_ERROR:
+        got_result = False
+        current_try = 0
+        
+        while not got_result and current_try<MAX_RETRIES-1:
+            current_try+=1
+            
+            try:
+                result = _get_req_session().request(
+                    method, request_url, params=params, files=files,
+                    timeout=(connect_timeout, read_timeout), proxies=proxy)
+                got_result = True
+                
+            except HTTPError:
+                logger.debug("HTTP Error on {0} method (Try #{1})".format(method_name, current_try))
+                time.sleep(RETRY_TIMEOUT)
+                
+            except ConnectionError:
+                logger.debug("Connection Error on {0} method (Try #{1})".format(method_name, current_try))
+                time.sleep(RETRY_TIMEOUT)
+                
+            except Timeout:
+                logger.debug("Timeout Error on {0} method (Try #{1})".format(method_name, current_try))
+                time.sleep(RETRY_TIMEOUT)
+            
+        
+        if not got_result:
+            result = _get_req_session().request(
+                    method, request_url, params=params, files=files,
+                    timeout=(connect_timeout, read_timeout), proxies=proxy)
+    else:
+        result = _get_req_session().request(
+            method, request_url, params=params, files=files,
+            timeout=(connect_timeout, read_timeout), proxies=proxy)
+    
     logger.debug("The server returned: '{0}'".format(result.text.encode('utf8')))
-    return _check_result(method_name, result)['result']
+    
+    json_result = _check_result(method_name, result)
+    if json_result:
+        return json_result['result']
 
 
 def _check_result(method_name, result):
@@ -82,23 +128,18 @@ def _check_result(method_name, result):
     :param result: The returned result of the method request
     :return: The result parsed to a JSON dictionary.
     """
-    if result.status_code != 200:
-        msg = 'The server returned HTTP {0} {1}. Response body:\n[{2}]' \
-            .format(result.status_code, result.reason, result.text.encode('utf8'))
-        raise ApiException(msg, method_name, result)
-
     try:
         result_json = result.json()
     except:
-        msg = 'The server returned an invalid JSON response. Response body:\n[{0}]' \
-            .format(result.text.encode('utf8'))
-        raise ApiException(msg, method_name, result)
-
-    if not result_json['ok']:
-        msg = 'Error code: {0} Description: {1}' \
-            .format(result_json['error_code'], result_json['description'])
-        raise ApiException(msg, method_name, result)
-    return result_json
+        if result.status_code != 200:
+            raise ApiHTTPException(method_name, result)
+        else:
+            raise ApiInvalidJSONException(method_name, result)
+    else:    
+        if not result_json['ok']:
+            raise ApiTelegramException(method_name, result, result_json)
+            
+        return result_json
 
 
 def get_me(token):
@@ -126,9 +167,8 @@ def download_file(token, file_path):
         
     result = _get_req_session().get(url, proxies=proxy)
     if result.status_code != 200:
-        msg = 'The server returned HTTP {0} {1}. Response body:\n[{2}]' \
-            .format(result.status_code, result.reason, result.text)
-        raise ApiException(msg, 'Download file', result)
+        raise ApiHTTPException('Download file', result)
+        
     return result.content
 
 
@@ -176,7 +216,7 @@ def set_webhook(token, url=None, certificate=None, max_connections=None, allowed
         files = {'certificate': certificate}
     if max_connections:
         payload['max_connections'] = max_connections
-    if allowed_updates:
+    if allowed_updates is not None:  # Empty lists should pass
         payload['allowed_updates'] = json.dumps(allowed_updates)
     return _make_request(token, method_url, params=payload, files=files)
 
@@ -192,7 +232,7 @@ def get_webhook_info(token):
     return _make_request(token, method_url, params=payload)
 
 
-def get_updates(token, offset=None, limit=None, timeout=None, allowed_updates=None):
+def get_updates(token, offset=None, limit=None, timeout=None, allowed_updates=None, long_polling_timeout = None):
     method_url = r'getUpdates'
     payload = {}
     if offset:
@@ -201,7 +241,9 @@ def get_updates(token, offset=None, limit=None, timeout=None, allowed_updates=No
         payload['limit'] = limit
     if timeout:
         payload['timeout'] = timeout
-    if allowed_updates:
+    if long_polling_timeout:
+        payload['long_polling_timeout'] = long_polling_timeout
+    if allowed_updates is not None:  # Empty lists should pass
         payload['allowed_updates'] = json.dumps(allowed_updates)
     return _make_request(token, method_url, params=payload)
 
@@ -296,10 +338,12 @@ def send_photo(
     method_url = r'sendPhoto'
     payload = {'chat_id': chat_id}
     files = None
-    if not util.is_string(photo):
-        files = {'photo': photo}
-    else:
+    if util.is_string(photo):
         payload['photo'] = photo
+    elif util.is_pil_image(photo):
+        files = {'photo': util.pil_image_to_file(photo)}
+    else:
+        files = {'photo': photo}
     if caption:
         payload['caption'] = caption
     if reply_to_message_id:
@@ -320,7 +364,7 @@ def send_media_group(
         disable_notification=None, reply_to_message_id=None,
         timeout=None):
     method_url = r'sendMediaGroup'
-    media_json, files = _convert_input_media_array(media)
+    media_json, files = convert_input_media_array(media)
     payload = {'chat_id': chat_id, 'media': media_json}
     if disable_notification is not None:
         payload['disable_notification'] = disable_notification
@@ -390,12 +434,14 @@ def stop_message_live_location(
 
 def send_venue(
         token, chat_id, latitude, longitude, title, address,
-        foursquare_id=None, disable_notification=None,
+        foursquare_id=None, foursquare_type=None, disable_notification=None,
         reply_to_message_id=None, reply_markup=None, timeout=None):
     method_url = r'sendVenue'
     payload = {'chat_id': chat_id, 'latitude': latitude, 'longitude': longitude, 'title': title, 'address': address}
     if foursquare_id:
         payload['foursquare_id'] = foursquare_id
+    if foursquare_type:
+        payload['foursquare_type'] = foursquare_type
     if disable_notification is not None:
         payload['disable_notification'] = disable_notification
     if reply_to_message_id:
@@ -408,13 +454,14 @@ def send_venue(
 
 
 def send_contact(
-        token, chat_id, phone_number, first_name,
-        last_name=None, disable_notification=None,
-        reply_to_message_id=None, reply_markup=None, timeout=None):
+        token, chat_id, phone_number, first_name, last_name=None, vcard=None,
+        disable_notification=None, reply_to_message_id=None, reply_markup=None, timeout=None):
     method_url = r'sendContact'
     payload = {'chat_id': chat_id, 'phone_number': phone_number, 'first_name': first_name}
     if last_name:
         payload['last_name'] = last_name
+    if vcard:
+        payload['vcard'] = vcard
     if disable_notification is not None:
         payload['disable_notification'] = disable_notification
     if reply_to_message_id:
@@ -435,7 +482,7 @@ def send_chat_action(token, chat_id, action, timeout=None):
 
 
 def send_video(token, chat_id, data, duration=None, caption=None, reply_to_message_id=None, reply_markup=None,
-               parse_mode=None, supports_streaming=None, disable_notification=None, timeout=None, thumb=None):
+               parse_mode=None, supports_streaming=None, disable_notification=None, timeout=None, thumb=None, width=None, height=None):
     method_url = r'sendVideo'
     payload = {'chat_id': chat_id}
     files = None
@@ -461,14 +508,21 @@ def send_video(token, chat_id, data, duration=None, caption=None, reply_to_messa
         payload['connect-timeout'] = timeout
     if thumb:
         if not util.is_string(thumb):
-            files['thumb'] = thumb
+            if files:
+                files['thumb'] = thumb
+            else:
+                files = {'thumb': thumb}
         else:
             payload['thumb'] = thumb
+    if width:
+        payload['width'] = width
+    if height:
+        payload['height'] = height
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
 def send_animation(token, chat_id, data, duration=None, caption=None, reply_to_message_id=None, reply_markup=None,
-               parse_mode=None, disable_notification=None, timeout=None):
+               parse_mode=None, disable_notification=None, timeout=None, thumb=None):
     method_url = r'sendAnimation'
     payload = {'chat_id': chat_id}
     files = None
@@ -490,6 +544,14 @@ def send_animation(token, chat_id, data, duration=None, caption=None, reply_to_m
         payload['disable_notification'] = disable_notification
     if timeout:
         payload['connect-timeout'] = timeout
+    if thumb:
+        if not util.is_string(thumb):
+            if files:
+                files['thumb'] = thumb
+            else:
+                files = {'thumb': thumb}
+        else:
+            payload['thumb'] = thumb
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
@@ -520,7 +582,7 @@ def send_voice(token, chat_id, voice, caption=None, duration=None, reply_to_mess
 
 
 def send_video_note(token, chat_id, data, duration=None, length=None, reply_to_message_id=None, reply_markup=None,
-                    disable_notification=None, timeout=None):
+                    disable_notification=None, timeout=None, thumb=None):
     method_url = r'sendVideoNote'
     payload = {'chat_id': chat_id}
     files = None
@@ -530,7 +592,7 @@ def send_video_note(token, chat_id, data, duration=None, length=None, reply_to_m
         payload['video_note'] = data
     if duration:
         payload['duration'] = duration
-    if length:
+    if length and (str(length).isdigit() and int(length) <= 639):
         payload['length'] = length
     else:
         payload['length'] = 639  # seems like it is MAX length size
@@ -542,6 +604,14 @@ def send_video_note(token, chat_id, data, duration=None, length=None, reply_to_m
         payload['disable_notification'] = disable_notification
     if timeout:
         payload['connect-timeout'] = timeout
+    if thumb:
+        if not util.is_string(thumb):
+            if files:
+                files['thumb'] = thumb
+            else:
+                files = {'thumb': thumb}
+        else:
+            payload['thumb'] = thumb
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
@@ -574,14 +644,17 @@ def send_audio(token, chat_id, audio, caption=None, duration=None, performer=Non
         payload['connect-timeout'] = timeout
     if thumb:
         if not util.is_string(thumb):
-            files['thumb'] = thumb
+            if files:
+                files['thumb'] = thumb
+            else:
+                files = {'thumb': thumb}
         else:
             payload['thumb'] = thumb
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
 def send_data(token, chat_id, data, data_type, reply_to_message_id=None, reply_markup=None, parse_mode=None,
-              disable_notification=None, timeout=None, caption=None):
+              disable_notification=None, timeout=None, caption=None, thumb=None):
     method_url = get_method_by_type(data_type)
     payload = {'chat_id': chat_id}
     files = None
@@ -601,6 +674,14 @@ def send_data(token, chat_id, data, data_type, reply_to_message_id=None, reply_m
         payload['connect-timeout'] = timeout
     if caption:
         payload['caption'] = caption
+    if thumb:
+        if not util.is_string(thumb):
+            if files:
+                files['thumb'] = thumb
+            else:
+                files = {'thumb': thumb}
+        else:
+            payload['thumb'] = thumb
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
@@ -619,9 +700,11 @@ def kick_chat_member(token, chat_id, user_id, until_date=None):
     return _make_request(token, method_url, params=payload, method='post')
 
 
-def unban_chat_member(token, chat_id, user_id):
+def unban_chat_member(token, chat_id, user_id, only_if_banned):
     method_url = 'unbanChatMember'
     payload = {'chat_id': chat_id, 'user_id': user_id}
+    if only_if_banned:
+        payload['only_if_banned'] = only_if_banned
     return _make_request(token, method_url, params=payload, method='post')
 
 
@@ -706,10 +789,12 @@ def set_chat_photo(token, chat_id, photo):
     method_url = 'setChatPhoto'
     payload = {'chat_id': chat_id}
     files = None
-    if not util.is_string(photo):
-        files = {'photo': photo}
-    else:
+    if util.is_string(photo):
         payload['photo'] = photo
+    elif util.is_pil_image(photo):
+        files = {'photo': util.pil_image_to_file(photo)}
+    else:
+        files = {'photo': photo}
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
@@ -791,7 +876,7 @@ def edit_message_caption(token, caption, chat_id=None, message_id=None, inline_m
 
 def edit_message_media(token, media, chat_id=None, message_id=None, inline_message_id=None, reply_markup=None):
     method_url = r'editMessageMedia'
-    media_json, file = _convert_input_media(media)
+    media_json, file = convert_input_media(media)
     payload = {'media': media_json}
     if chat_id:
         payload['chat_id'] = chat_id
@@ -818,9 +903,11 @@ def edit_message_reply_markup(token, chat_id=None, message_id=None, inline_messa
     return _make_request(token, method_url, params=payload, method='post')
 
 
-def delete_message(token, chat_id, message_id):
+def delete_message(token, chat_id, message_id, timeout=None):
     method_url = r'deleteMessage'
     payload = {'chat_id': chat_id, 'message_id': message_id}
+    if timeout:
+        payload['connect-timeout'] = timeout
     return _make_request(token, method_url, params=payload, method='post')
 
 
@@ -900,7 +987,8 @@ def get_game_high_scores(token, user_id, chat_id=None, message_id=None, inline_m
 def send_invoice(
         token, chat_id, title, description, invoice_payload, provider_token, currency, prices,
         start_parameter, photo_url=None, photo_size=None, photo_width=None, photo_height=None,
-        need_name=None, need_phone_number=None, need_email=None, need_shipping_address=None, is_flexible=None,
+        need_name=None, need_phone_number=None, need_email=None, need_shipping_address=None,
+        send_phone_number_to_provider = None, send_email_to_provider = None, is_flexible=None,
         disable_notification=None, reply_to_message_id=None, reply_markup=None, provider_data=None,
         timeout=None):
     """
@@ -923,10 +1011,13 @@ def send_invoice(
     :param need_email: Pass True, if you require the user's email to complete the order
     :param need_shipping_address: Pass True, if you require the user's shipping address to complete the order
     :param is_flexible: Pass True, if the final price depends on the shipping method
+    :param send_phone_number_to_provider: Pass True, if user's phone number should be sent to provider
+    :param send_email_to_provider: Pass True, if user's email address should be sent to provider
     :param disable_notification: Sends the message silently. Users will receive a notification with no sound.
     :param reply_to_message_id: If the message is a reply, ID of the original message
     :param reply_markup: A JSON-serialized object for an inline keyboard. If empty, one 'Pay total price' button will be shown. If not empty, the first button must be a Pay button
-    :param provider_data:
+    :param provider_data: A JSON-serialized data about the invoice, which will be shared with the payment provider. A detailed description of required fields should be provided by the payment provider.
+    :param timeout:
     :return:
     """
     method_url = r'sendInvoice'
@@ -949,6 +1040,10 @@ def send_invoice(
         payload['need_email'] = need_email
     if need_shipping_address is not None:
         payload['need_shipping_address'] = need_shipping_address
+    if send_phone_number_to_provider is not None:
+        payload['send_phone_number_to_provider'] = send_phone_number_to_provider
+    if send_email_to_provider is not None:
+        payload['send_email_to_provider'] = send_email_to_provider
     if is_flexible is not None:
         payload['is_flexible'] = is_flexible
     if disable_notification is not None:
@@ -1164,13 +1259,13 @@ def _convert_markup(markup):
     return markup
 
 
-def _convert_input_media(media):
+def convert_input_media(media):
     if isinstance(media, types.InputMedia):
-        return media._convert_input_media()
+        return media.convert_input_media()
     return None, None
 
 
-def _convert_input_media_array(array):
+def convert_input_media_array(array):
     media = []
     files = {}
     for input_media in array:
@@ -1195,7 +1290,7 @@ def _no_encode(func):
 
 class ApiException(Exception):
     """
-    This class represents an Exception thrown when a call to the Telegram API fails.
+    This class represents a base Exception thrown when a call to the Telegram API fails.
     In addition to an informative message, it has a `function_name` and a `result` attribute, which respectively
     contain the name of the failed function and the returned result that made the function to be considered  as
     failed.
@@ -1205,3 +1300,41 @@ class ApiException(Exception):
         super(ApiException, self).__init__("A request to the Telegram API was unsuccessful. {0}".format(msg))
         self.function_name = function_name
         self.result = result
+    
+class ApiHTTPException(ApiException):
+    """
+    This class represents an Exception thrown when a call to the 
+    Telegram API server returns HTTP code that is not 200.
+    """
+    def __init__(self, function_name, result):
+        super(ApiHTTPException, self).__init__(
+            "The server returned HTTP {0} {1}. Response body:\n[{2}]" \
+            .format(result.status_code, result.reason, result.text.encode('utf8')),
+            function_name,
+            result)
+    
+class ApiInvalidJSONException(ApiException):
+    """
+    This class represents an Exception thrown when a call to the 
+    Telegram API server returns invalid json.
+    """
+    def __init__(self, function_name, result):
+        super(ApiInvalidJSONException, self).__init__(
+            "The server returned an invalid JSON response. Response body:\n[{0}]" \
+            .format(result.text.encode('utf8')),
+            function_name,
+            result)
+    
+class ApiTelegramException(ApiException):
+    """
+    This class represents an Exception thrown when a Telegram API returns error code.
+    """
+    def __init__(self, function_name, result, result_json):
+        super(ApiTelegramException, self).__init__(
+            "Error code: {0}. Description: {1}" \
+            .format(result_json['error_code'], result_json['description']),
+            function_name,
+            result)
+        self.result_json = result_json
+        self.error_code = result_json['error_code']
+        
